@@ -1046,7 +1046,7 @@ struct oraTable
 void
 oracleEstimate(oracleSession *session, const char *query, double seq_page_cost, int block_size, double *startup_cost, double *total_cost, double *rows, int *width)
 {
-	double time;
+	double time, bytes;
 	dvoid *res[3];
 	sb4 res_size[3];
 	ub2 res_type[3], res_len[3];
@@ -1061,15 +1061,18 @@ oracleEstimate(oracleSession *session, const char *query, double seq_page_cost,
 	res_size[1] = sizeof(double);
 	res_type[1] = SQLT_BDOUBLE;
 
-	res[2] = (dvoid *)width;
-	res_size[2] = sizeof(int);
-	res_type[2] = SQLT_INT;
+	res[2] = (dvoid *)&bytes;
+	res_size[2] = sizeof(double);
+	res_type[2] = SQLT_BDOUBLE;
 
 	oracleQueryPlan(session, query, desc_query, 3, res, res_size, res_type, res_len, res_ind);
 
 	/* close the statement */
 	oracleCloseStatement(session);
 
+	/* width now contains the total bytes estimated - divide by number of rows */
+	*width = (int)(bytes / *rows);
+
 	/*
 	 * Guess startup_cost and total_cost from Oracle's "time".
 	 * This is really shady as Oracle's time has a granularity of 1.
@@ -1740,13 +1743,14 @@ oracleFetchNext(oracleSession *session, struct oraTable *oraTable)
  * oracleGetLob
  * 		Get the LOB contents and store them in *value and *value_len.
  * 		"oraTable" is passed so that all LOB locators can be freed in case of errors.
+ * 		If "trunc" is nonzero, it contains the number of bytes or characters to get.
  */
 void
-oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, oraType type, char **value, long *value_len)
+oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, oraType type, char **value, long *value_len, unsigned long trunc)
 {
 	OCILobLocator *locp = *(OCILobLocator **)locptr;
-	ub4 amount;
-	sword result;
+	oraub8 amount_byte, amount_char;
+	sword result = OCI_SUCCESS;
 
 	/* initialize result buffer length */
 	*value_len = 0;
@@ -1776,14 +1780,17 @@ oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, or
 			*value = oracleRealloc(*value, *value_len + LOB_CHUNK_SIZE + 1);
 
 		/*
-		 * The first time round, "amount = 0" tells OCILobRead to read the whole LOB.
-		 * On subsequent reads, the parameter is ignored.
-		 * After the call, "amount" contains the number of bytes read.
+		 * The first time round, "amount_* = 0" tells OCILobRead to read the whole LOB.
+		 * On subsequent reads, the amount_* parameters are ignored.
+		 * After the call, "amount_byte" contains the number of bytes read.
 		 */
-		amount = 0;
+		amount_byte = (oraub8)trunc;  /* ignored for CLOBs */
+		amount_char = amount_byte;  /* ignored for binary LOBs */
 		result = checkerr(
-			OCILobRead(session->connp->svchp, session->envp->errhp, locp, &amount, (ub4)1,
-				(dvoid *)(*value + *value_len), LOB_CHUNK_SIZE, NULL, NULL, (ub2)0, 0),
+			OCILobRead2(session->connp->svchp, session->envp->errhp, locp, &amount_byte, &amount_char,
+				(oraub8)1, (dvoid *)(*value + *value_len), (oraub8)LOB_CHUNK_SIZE,
+				(result == OCI_NEED_DATA) ? OCI_NEXT_PIECE : OCI_FIRST_PIECE,
+				NULL, NULL, (ub2)0, (ub1)0),
 			(dvoid *)session->envp->errhp, OCI_HTYPE_ERROR);
 
 		if (result == OCI_ERROR)
@@ -1794,7 +1801,7 @@ oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, or
 		}
 
 		/* update LOB length */
-		*value_len += amount;
+		*value_len += amount_byte;
 	}
 	while (result == OCI_NEED_DATA);
 
